//////////////////////////////////////////////////////////////////////
//	Kustin--Miller, Papadakis--Reid unprojection with sequences
//////////////////////////////////////////////////////////////////////

// Note: include this file in magma spec file or do Attach("unproj.m");

forward km;
intrinsic KustinMillerLifting( IX::RngMPol, ID::RngMPol ) -> RngMPol,Map
{}

	// Step 0: check inputs and prepare the free resolutions
	R := Generic(ID);
	assert Generic(IX) eq R;
	assert IX subset ID;
	MD := quo< Module(R,1) | [ [f] : f in Basis(ID) ] >;
	MX := quo< Module(R,1) | [ [f] : f in Basis(IX) ] >;
	FD := FreeResolution(MD);
	FX := FreeResolution(MX);
	assert #BoundaryMaps(FD) eq #BoundaryMaps(FX) + 1;

	// Prepare a ring to hold the solution.
	n := Rank(R);
	S := PolynomialRing( BaseRing(R), n + 1 );
	RtoS := hom< R -> S | [ S.i : i in [1..n] ] >;

	return km(FX, FD, IX, RtoS), RtoS;
end intrinsic;

intrinsic KustinMillerLifting( IX::RngMPol, ID::RngMPol, S::RngMPol )
			-> RngMPol,Map
{}
	// Step 0: check inputs and prepare the free resolutions
	R := Generic(ID);
	assert Generic(IX) eq R;
	assert IX subset ID;
	MD := quo< Module(R,1) | [ [f] : f in Basis(ID) ] >;
	MX := quo< Module(R,1) | [ [f] : f in Basis(IX) ] >;
	FD := FreeResolution(MD);
	FX := FreeResolution(MX);
	assert #BoundaryMaps(FD) eq #BoundaryMaps(FX) + 1;

	n := Rank(R);
	RtoS := hom< R -> S | [ S.i : i in [1..n] ] >;

	return km(FX, FD, IX, RtoS), RtoS;
end intrinsic;

intrinsic KustinMillerLifting( IX::RngMPol, ID::RngMPol, f::Map )
			-> RngMPol,Map
{The Kustin--Miller and Papadakis--Reid unprojection of the ideals IX and ID
of a common polynomial ring R. The ideals must satisfy: (1) ID, IX are
Gorenstein; (2) ID contains IX; (3) depth(ID) = depth(IX) + 1.
If a homomorphism f from R to a  suitable polynomial ring S is supplied
(or even just S alone if the map is the inclusion to the first variables),
the result is returned as an ideal of S. Otherwise a new
ring S and map f: R -> S is created to hold the result. The map f: R -> S 
is returned as a second value.}

	// Step 0: check inputs and prepare the free resolutions
	R := Generic(ID);
	assert Generic(IX) eq R;
	assert IX subset ID;
	MD := quo< Module(R,1) | [ [f] : f in Basis(ID) ] >;
	MX := quo< Module(R,1) | [ [f] : f in Basis(IX) ] >;
	FD := FreeResolution(MD);
	FX := FreeResolution(MX);
	assert #BoundaryMaps(FD) eq #BoundaryMaps(FX) + 1;

	return km(FX, FD, IX, f);
end intrinsic;

// I and J two ideals in the same polynomial ring.
// Conditions:
//	X = V(J) contains D = V(I)
//		i.e. J is a subset of I.
//	D in X is pure codimension 1
//		i.e. free res of R/I is one step longer than R/J.

function km( FX, FD, IX, RtoS  )
	R := Generic(IX);
	n := Rank(R);
	mapsD := BoundaryMaps(FD);
	mapsX := BoundaryMaps(FX);
	k := #mapsX - 1;

	Phi := [* *];	// this will hold the matrices that make up the
		// map of complexes - can't be seq since mxs have diff sizes.

	// Step 1: lift the identify map through the complexes
	phi := IdentityMatrix(R,1);
	for i in [k..2 by -1] do
		phi := Solution( Matrix(mapsD[i+1]), Matrix(mapsX[i]) * phi );
		Append(~Phi, Matrix(phi));
	end for;

	// Step 2: build the unprojection ring
	S := Codomain(RtoS);
	phi_k := RtoS(Eltseq(Phi[#Phi]));
	s_D := RtoS(Eltseq(Matrix(mapsD[2])));

	unprojeqns := [ S.(n+1) * s_D[i] -  phi_k[i] : i in [1..#s_D] ];
	K := ideal< S | RtoS(Basis(IX)), unprojeqns >;
	return K;
end function;

/*	Examples

// Example 1: D = c.i.2 in X = hypersurface

k := Rationals();
R<x,y,A,B> := PolynomialRing(k,4);
ID := Ideal([x,y]);
IX := Ideal([x*A - y*B]);
K,phi := KustinMillerLifting(IX,ID);	// phi maps R to unprojection ring
_<x,y,A,B,s> := Generic(K);	// just to set some variable names
K;
	Ideal of Polynomial ring of rank 5 over Rational Field
	Order: Lexicographical
	Variables: x, y, A, B, s
	Basis:
	[
	    x*A - y*B,
	    -y*s - A,
	    x*s + B
	]

// Example 2: D = c.i.3 in X = c.i.2

k := Rationals();
R<x,y,z,u,v,w,a,b,c> := PolynomialRing(k,9);
ID := Ideal([a,b,c]);
M := Matrix(2,3,[x,y,z,u,v,w]);
A := Matrix(3,1,[a,b,c]);
IX := Ideal(Eltseq(M*A));
K,phi := KustinMillerLifting(IX,ID);
_<[x]> := Generic(K);	// set some variable names
K;
	Ideal of Polynomial ring of rank 10 over Rational Field
	Order: Lexicographical
	Variables: x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9], x[10]
	Homogeneous
	Basis:
	[
	    x[1]*x[7] + x[2]*x[8] + x[3]*x[9],
	    x[4]*x[7] + x[5]*x[8] + x[6]*x[9],
	    -x[1]*x[5] + x[2]*x[4] + x[9]*x[10],
	    -x[1]*x[6] + x[3]*x[4] - x[8]*x[10],
	    -x[2]*x[6] + x[3]*x[5] + x[7]*x[10]
	]

// Example 3: D = c.i.4 in X = c.i.3 5x5 Pfaffian - also set the unproj ring S.

k := Rationals();
R<b,c,d,e,f,L,M> := PolynomialRing(k, [2,3,4,4,5,4,5]);
ID := Ideal([c,d,e,f]);
PfMx := -AntisymmetricMatrix([ R | 0, b, c, c, d, e, L, f, M, 0 ]);
IX := Ideal(Pfaffians(PfMx,4));
S<b,c,d,e,f,L,M,g> := PolynomialRing(k, [2,3,4,4,5,4,5,3]);
phi := hom< R -> S | [ S.i : i in [1..Rank(R)] ] >;
K := KustinMillerLifting(IX,ID,phi); K;
*/

