//////////////////////////////////////////////////////////////////////
// Tom and Jerry matrices
// Gavin Brown, Michael Kerber, Miles Reid, 2005-10
//
//  Contains constructors for random Tom and Jerry matrices.
//
//////////////////////////////////////////////////////////////////////


forward RandPoly, RandEltIdealD,
	RowColumnOperationsTom, RowColumnOperationsJerry;

/* Test example: build two matrices and assert what should work in each case.

wts := -AntisymmetricMatrix([2,3,4,4,5,6,5,6,7,8]);
K := FiniteField(65521);
R<x1,x2,x3,y1,y2,y3,y4> := PolynomialRing(K,[1,2,3,4,5,5,6]);
D := [y1,y2,y3,y4];
M := TomMatrix(wts,3,D);
assert &and[ M[i,j] in Ideal(D) : j in [i+1..5], i in [1..4] | not 3 in {i,j} ];
assert M + Transpose(M) eq ZeroMatrix(R,5,5);
assert M[1,2] eq 0 and M[4,5] eq 0;

wts := -AntisymmetricMatrix([3,4,5,5,6,7,6,7,8,9]);
K := FiniteField(65521);
R<x1,x2,x3,y1,y2,y3,y4> := PolynomialRing(K,[1,2,3,4,5,7,9]);
D := [y1,y2,y3,y4];
M := JerryMatrix(wts,3,4,D);
assert &and [ M[i,j] in Ideal(D) :
	 j in [i+1..5], i in [1..4] | 3 in {i,j} or 4 in {i,j} ];
assert M + Transpose(M) eq ZeroMatrix(R,5,5);
assert M[3,5] eq 0;

*/

//////////////////////////////////////////////////////////////////////
//              Tom and Jerry matrix constructors
//////////////////////////////////////////////////////////////////////

intrinsic TomMatrix(wts::Mtrx,i0::RngIntElt,GensID::SeqEnum) -> Mtrx
{A graded 5x5 antisymmetric matrix in Tom(i0) format with gradings
given by the upper triangular entries of the integer matrix wts and
Tom ideal generated by the (graded) polynomial ring variables in the
sequence GensID}
    RR := Universe(GensID);
    W := Grading(RR);
    // Step 1: Build a skew 5x5 matrix in Tom(i0) format with weights W and
    // unproj'n ideal gens the vars of indices G det'g the unprojection plane D.
    copyofGensID := GensID;
    othervars := [ RR.i : i in [1..Rank(RR)] | not RR.i in GensID ];
    M := ZeroMatrix(RR,5,5);
    for i in [1..4] do
        for j in [i+1..5] do
            degij := wts[i,j];
            if #{i,j,i0} eq 3 then
                // neither i nor j equals i0, so we should be in ID
                canusevariable := Index(
                    [WeightedDegree(g):g in copyofGensID],degij);
                if canusevariable ne 0 then
                    // we can use this generator in this matrix position
                    F := copyofGensID[canusevariable];
                    Remove(~copyofGensID,canusevariable);
                else
                    F := RandEltIdealD(degij,GensID);
                end if;
            else
                canusevariable:=Index(
                    [WeightedDegree(g):g in othervars],degij);
                if canusevariable ne 0 then
                    // we can use this generator in this matrix position
                    F := othervars[canusevariable];
                    Remove(~othervars,canusevariable);
                else
                    F := RandPoly(degij,RR);
                end if;
            end if;
            M[i,j] := F;
            M[j,i] := -F;
        end for;
    end for;
    RowColumnOperationsTom(~M,GensID,i0);
    return M;
end intrinsic;

intrinsic JerryMatrix(wts::Mtrx,i0,j0::RngIntElt,GensID::SeqEnum) -> Mtrx
{A graded 5x5 antisymmetric matrix in Jerry(i0,j0) format with gradings
given by the upper triangular entries of the integer matrix wts and
Tom ideal generated by the (graded) polynomial ring variables in the
sequence GensID}
    RR := Universe(GensID);
    W := Grading(RR);
    x := RR.1;
    // Step 1: Build a skew 5x5 matrix in Jerry(i0,j0) format with weights W and
    // unproj'n ideal gens the vars of indices G det'g the unprojection plane D.
    copyofGensID := GensID;
    othervars := [ RR.i : i in [1..Rank(RR)] | not RR.i in GensID ];
    M := ZeroMatrix(RR,5,5);
    for i in [1..4] do
        for j in [i+1..5] do
            degij := wts[i,j];
            if #{i,j,i0,j0} le 3 then // at least one of i,j equals i0,j0
                canusevariable := Index(
                    [WeightedDegree(g):g in copyofGensID],degij);
                if canusevariable ne 0 then
                    // we can use this generator in this matrix position
                    F := copyofGensID[canusevariable];
                    Remove(~copyofGensID,canusevariable);
                else
                    F := RandEltIdealD(degij,GensID);
                end if;
            else
                canusevariable := Index(
                    [WeightedDegree(g):g in othervars],degij);
                if canusevariable ne 0 then
                    // we can use this generator in this matrix position
                    F := othervars[canusevariable];
                    Remove(~othervars,canusevariable);
                else
                    F := RandPoly(degij,RR);
                end if;
            end if;
            M[i,j] := F;
            M[j,i] := -F;
        end for;
    end for;
    RowColumnOperationsJerry(~M,GensID,i0,j0);
    return M;
end intrinsic;


//////////////////////////////////////////////////////////////////////
//		Auxilliary functions
//////////////////////////////////////////////////////////////////////

// A random poly of weighted degree dd in the graded poly ring RR.
function RandPoly(dd,RR)
        if dd ge 0 then
                return &+[ RR | Random(CoefficientRing(RR))*m :
                        m in MonomialsOfWeightedDegree(RR,dd)];
        else
                return RR!0;
        end if;
end function;

// A random poly of weighted degree dd in the graded ideal gen'd by GensID.
function RandEltIdealD(dd,GensID)
	RR := Universe(GensID);
	return &+[ RR | RandPoly(dd-WeightedDegree(m),RR)*m : m in GensID];
end function;

forward cancel_entry;

    // Make some attempt at row/col ops in easy cases - might create a zero
    // that would be useful later.
    // If we find an entry in the Tom zone that is a single variable,
    // then we look later (in same row/col) for entries that are divisible
    // by it. Whenever we find one, we use row/col ops to remove it.
    // (Single var entries exist by the 'canusevariable' mechanism.)
    // For Tom(i), any row/col op Row(j) :-> Row(j) - (any)*Row(k) is fine
    // for any j,k ne i (in that it doesn't break the Tom(i) format).
    // Note that this does not change the ideal, so the gb calculations
    // still take time - but it may identify a reason for failure sooner.
    // N.B. This works in row/col order, and does not eliminate earlier entries.

procedure RowColumnOperationsTom(~M,GensID,i0)
    ops_done := { Parent({1,2}) | };  // record row/col ops already spent.
    for i in [1..4] do
        if i eq i0 then continue i; end if;
	for j in [i+1..5] do
	    if j eq i0 then continue j; end if;
	    monos_ij := Monomials(M[i,j]);
	    if #monos_ij eq 1 and monos_ij[1] in GensID then
	        x := monos_ij[1];
		// Attempt to remove multiples of x from later rows/cols.
		for k in [i..5] do
		    for l in [Maximum([j,k+1])..5] do
			// ignore this case if we're not in the Jerry zone.
			if i0 in {k,l} then continue l; end if;
		    	if #({i,j} meet {k,l}) ne 1 then continue l; end if;
			cancel_entry(~M,~ops_done,x,i,j,k,l);
		    end for;
		end for;
	    end if;
	end for;
    end for;
end procedure;

// If we find an entry in the Jerry zone that is a single variable,
// then we look later (in same row/col) for entries that are divisible
// by it. Whenever we find one, we use row/col ops to remove it.
// (Single var entries exist by the 'canusevariable' mechanism.)
// For Jerry(i,j), any row/col op Row(k) :-> Row(k) - (any)*Row(l) is fine
// for any j,k ne i (in that it doesn't break the Tom(i) format).
// Note that this does not change the ideal, so the gb calculations
// still take time - but it may identify a reason for failure sooner.
// N.B. This works in row/col order, and does not eliminate earlier entries.

procedure RowColumnOperationsJerry(~M,GensID,i0,j0)
    ops_done := { Parent({1,2}) | };  // record row/col ops already spent.
    for i in [1..4] do
	for j in [i+1..5] do
	    // ignore this case if we're not in the Jerry zone.
	    if #({i,j} meet {i0,j0}) eq 0 then continue j; end if;
	    monos_ij := Monomials(M[i,j]);
	    if #monos_ij eq 1 and monos_ij[1] in GensID then
	        x := monos_ij[1];
		// Attempt to remove multiples of x from later rows/cols.
		for k in [i..5] do
		    for l in [Maximum([j,k+1])..5] do
			// ignore this case if we're not in the Jerry zone.
			if #({k,l} meet {i0,j0}) eq 0 then continue k; end if;
		    	if #({i,j} meet {k,l}) ne 1 then continue l; end if;
			cancel_entry(~M,~ops_done,x,i,j,k,l);
		    end for;
		end for;
	    end if;
	end for;
    end for;
end procedure;

// We have M[i,j] = x (up to a scalar factor) and we want to use a row-col
// operation to remove multiples of x from M[k,l] as long as that
// row-col operation has not already been used (i.e. is not in ops_done).
// We assume i <= k and j <= l (which is fine if matrix wts are increasing)
// and also i < j and k < l (i.e. only work on upper triangle of M).
// (In ptic, i <= k < l.)

procedure cancel_entry(~M,~ops_done,x,i,j,k,l);
    xcoeff := Coefficient(M[i,j],x,1);
    divbleterms := &+[Parent(x)| t : t in Terms(M[k,l]) | IsDivisibleBy(t,x) ];
    coeff := divbleterms div x;
    // find the rowcol op i1->j1 we want (i.e. Row(j1) -> Row(j1) - ..*Row(i1))
    i1 := i eq k select j else i;
    j1 := k in {i,j} select l else k;
    if not ({i1,j1} in ops_done) and coeff ne 0 then
        rowcol := i1 eq i select j else i;
        sign := rowcol lt Minimum([i1,j1]) or rowcol gt Maximum([i1,j1])
    	    select -1 else 1;
        AddRow(~M,sign*coeff/xcoeff,i1,j1);
        AddColumn(~M,sign*coeff/xcoeff,i1,j1);
	Include(~ops_done, {i1,j1});
    end if;
end procedure;

