//////////////////////////////////////////////////////////////////////
// Tom and Jerry codimension 3 formats and their unprojections
// Gavin Brown, Michael Kerber, Miles Reid, 2005-10
//
//  Constructor for codim 3 numerical candidate.
//	 TomAndJerryCodimension3( W, u, w ) -> record of Y
//
//  Constructor for projection from codim 4.
//	 TomAndJerryUnprojection( Y, E, n ) -> record of Tom(n) for E subset Y
//	 TomAndJerryUnprojection( Y, E, n, m ) -> record of Jerry(n,m)
//	 TomAndJerryUnprojections( Y, E ) -> seq of all Tom and Jerry records
//
//		Notation
//		--------
//  We work with projections of the form
//	P in X in P(W) - -> E in Y in P(W omit r)
//  where
//	X	codim 4 Fano 3-fold
//	P in X	Type I centre on X of type 1/r(1,a,r-a)
//		(Type I means that r, 1, a, r-a appear separately in W)
//	Y	codim 3 - syzygy matrix determined by integer u and 5-tuple w
//	E in Y	the exceptional locus of the projection.
//
//////////////////////////////////////////////////////////////////////

import "recordtj.m":
	FanoRecord,
	TomJerryRecord;
import "failuretj.m":
	does_each_variable_appear,
	point_of_embdim6,
	non_terminal_coord_point,
	singular_points;

//////////////////////////////////////////////////////////////////////
//		Constructor for making codim 3 numerical candidates
//////////////////////////////////////////////////////////////////////

forward tj_equivalences, build_unprojection;

intrinsic TomAndJerryCodimension3(W::SeqEnum,u::RngIntElt,w::SeqEnum) -> Rec
{A record of a codimension 3 variety in 5x5 Pfaffian format, where W is
a sequence of integers (the ambient weights), u is an integer and w is
a sequence of five integers}
	require Universe(w) eq Integers() and #w eq 5:
		"The sequence w must consist of 5 integers";
	X := rec< FanoRecord | >;
	X`weights := Sort(W);
	X`codim := 3;
	X`u := u;
	X`w := w;
	X`WM := Matrix(5,5,[ u+w[i]+w[j] : i in [1..5],j in [1..5]]);
	X`tjequiv := tj_equivalences(w);
	X`k := 5*u + 2*&+w - &+W;
	return X;
end intrinsic;


//////////////////////////////////////////////////////////////////////
//		Constructor for making unprojections
//////////////////////////////////////////////////////////////////////

intrinsic TomAndJerryUnprojection(Y::Rec, E::SeqEnum, n::RngIntElt :
		 q := 65521, all_tests := true ) -> Rec
{Attempt Tom(n) unprojection of divisor P(E), where E is a sequence of
integers, of the codimension 3 variety described by the record Y}
	require Y`codim eq 3: "Y must be a record of a codim 3 variety";
	require n in [1..5]: "Tom type n must be an integer in the range 1..5";
	Sort(~E);
	W := Y`weights;
	// We need E subseq W and of the right codim 4 length.
	require IsSubsequence( E, W : Kind:="Sequential" ) and #E eq #W - 4:
		"Unprojection weights E must be a subsequence of ambient" *
			" weights W of length 4 less"; 
	return build_unprojection(Y,E,"Tom",[n] :
			q := q, all_tests := all_tests );
end intrinsic;


intrinsic TomAndJerryUnprojection(Y::Rec, E::SeqEnum, n,m::RngIntElt :
			 q := 65521, all_tests := true ) -> Rec
{Attempt Jerry(n,m) unprojection of divisor P(E), where E is a sequence of
integers, of the codimension 3 variety described by the record Y}
	require Y`codim eq 3: "Y must be a record of a codim 3 variety";
	require n in [1..5] and m in [1..5] and n ne m:
		"Jerry type (n,m) must be distinct integers in the range 1..5";
	Sort(~E);
	W := Y`weights;
	// We need E subseq W and of the right codim 4 length.
	require IsSubsequence( E, W : Kind:="Sequential" ) and #E eq #W - 4:
		"Unprojection weights E must be a subsequence of ambient" *
			" weights W of length 4 less"; 
	return build_unprojection(Y,E,"Jerry",[n,m] :
			q := q, all_tests := all_tests );
end intrinsic;

intrinsic TomAndJerryUnprojections(Y::Rec, E::SeqEnum :
		 q := 65521, all_tests := true ) -> Rec
{Attempt all Tom and Jerry unprojections of divisor P(E), where E is a
sequence of integers, of the codimension 3 variety described by the record Y}
	require Y`codim eq 3: "Y must be a record of a codim 3 variety";
	Sort(~E);
	W := Y`weights;
	// We need E subseq W and of the right codim 4 length.
	require IsSubsequence( E, W : Kind:="Sequential" ) and #E eq #W - 4:
		"Unprojection weights E must be a subsequence of ambient" *
			" weights W of length 4 less"; 

	// Recover inequivalent Tom and Jerry formats from Y
        toms := [StringToInteger(p[1][5]) :p in Y`tjequiv |p[1][1..3] eq "Tom"];
        jerries := [ [StringToInteger(p[1][7]), StringToInteger(p[1][9])]
                  : p in Y`tjequiv | p[1][1..5] eq "Jerry" ];

	// Check them all
	unprojections := [];
	for n in toms do
		p := build_unprojection(Y,E,"Tom",[n] :
			q := q, all_tests := all_tests );
		if GetVerbose("User1") eq 1 then TomAndJerryPrint(p); end if;
		Append(~unprojections, p);
	end for;
	for nm in jerries do
		p := build_unprojection(Y,E,"Jerry",nm :
			q := q, all_tests := all_tests );
		if GetVerbose("User1") eq 1 then TomAndJerryPrint(p); end if;
		Append(~unprojections, p);
	end for;
	return unprojections;
end intrinsic;


//////////////////////////////////////////////////////////////////////
//		Build a specific unprojection
//////////////////////////////////////////////////////////////////////

forward fishy_zero;
function build_unprojection(Y,E,type,number : q := 65521, all_tests := true )
	// We need to fix a subseq E of W.
	W := Y`weights;
	Wdelete := W;
	Eindices := [];
	for i in [1..#E] do
		ind := Index(Wdelete,E[i]);	// not 0 since E subseq W.
		Append(~Eindices, ind + i - 1);	// to account for removals.
		Remove(~Wdelete,ind);
	end for;
	// record the indices of the ideal defining this E in W.
	Dindices := [ i : i in [1..#W] | not i in Eindices ];
	pi := rec< TomJerryRecord | >;
	pi`success := true;	// optimistic start.
	if type eq "Tom" then
		pi`name := "Tom(" cat IntegerToString(number[1]) cat ")";
	else
		pi`name := "Jerry(" cat IntegerToString(number[1]) cat
					IntegerToString(number[2]) cat ")";
	end if;
	pi`r := &+ E + Y`k;
	pi`A := E;
	pi`Dindices := Dindices;

	// Build a Tom(1) matrix and the equations of Y.
        KK := FiniteField(q);
        RR<[x]> := PolynomialRing(KK, W);
        GensID := [ RR.i : i in [1..#W] | i in Dindices ];

	WM := Y`WM;	// duplication of data - not a big deal.
	pi`WM := WM;
	pi`WY := Y`weights;
	pi`u := Y`u;
	pi`w := Y`w;
	if type eq "Tom" then
		M := TomMatrix(WM,number[1],GensID);
	else
		M := JerryMatrix(WM,number[1],number[2],GensID);
	end if;
	pi`M := M;

	zeroes := fishy_zero(M);	// a seq [ [i,j] ... ] of zeros in M.
	pi`fishy_zeroes := zeroes;
	NZ := #zeroes;
	if NZ ne 0 then
		// collinear zeros is an easy fail
		if #{ z[1] : z in zeroes} ne NZ
				or #{ z[2] : z in zeroes} ne NZ then
			pi`collinear_zeroes := true;
			pi`success := false;
			if not all_tests then
				return pi;
			end if;
		end if;
		// the scheme def'd by the row/cols through any 0 is in Sing(Y)
		for z in zeroes do
			Z := Ideal([ M[z[1],j] : j in [1..5] | j ne z[1] ] cat
				[ M[i,z[2]] : i in [1..5] | i ne z[2] ]);
			if Dimension(Z) ge 2 then	// i.e. proj dim >= 1
				pi`bad_zero := true;
				pi`success := false;
				if not all_tests then
					return pi;
				end if;
			end if;
		end for;
	end if;

	// Now do the tests
	IY := Ideal(Pfaffians(M,4));
	variable_appears := does_each_variable_appear(IY);
	pi`variable_appears := variable_appears;
	if not &and variable_appears then
		pi`success := false;
		if not all_tests then
			return pi;
		end if;
	end if;

	pure_power_appears := point_of_embdim6(M);
	pi`pure_power_appears := pure_power_appears;
	if not &and pure_power_appears then
		pi`success := false;
		if not all_tests then
			return pi;
		end if;
	end if;

	// Check 0-strata off D for obvious non-terminal points.
	non_mori_pt_off_D := non_terminal_coord_point(Basis(IY),[i : i in [1..7] | not i in Eindices]);
	pi`non_mori_pt_off_D := non_mori_pt_off_D;
	if non_mori_pt_off_D then
		pi`success := false;
		if not all_tests then
			return pi;
		end if;
	end if;

	// Check 0-strata on D for obvious non-terminal points.
	non_mori_pt_on_D := non_terminal_coord_point(Basis(IY),Eindices);
	pi`non_mori_pt_on_D := non_mori_pt_on_D;
	if non_mori_pt_on_D then
		pi`success := false;
		if not all_tests then
			return pi;
		end if;
	end if;

	are_sings_on_D,sings_are_red,sings_on_pts,sings_on_lines,num_sings :=
		singular_points(M, GensID : all_tests := true);
	works := &and pure_power_appears and are_sings_on_D and sings_are_red;
	pi`success and:= works;
	pi`are_sings_on_D := are_sings_on_D;
	if assigned sings_are_red then
		pi`sing_is_red := sings_are_red;
	end if;
	if assigned sings_on_pts then
		pi`are_0_strata_singular := sings_on_pts;
	end if;
	if assigned sings_on_lines then
		pi`are_1_strata_singular := sings_on_lines;
	end if;
	if assigned num_sings then
		pi`N := num_sings;
	end if;

	return pi;
end function;



//////////////////////////////////////////////////////////////////////
//		Auxilliary functions
//////////////////////////////////////////////////////////////////////

// A function to decide when different Tom or Jerry unprojections
// are equivalent by row/column operations on the 5x5 syzygy matrix.
// This only does trivial numerology: when do 2 rows have the same weights.
function tj_equivalences(w)
	tjequiv := [ ];
	wred := Setseq(Seqset(w));
	// Detect equivalent Tom formats
	toms := [ Parent(["Tom(1)","Tom(2))"]) | [] : i in [1..#wred] ];
	for i in [1..5] do
		position := Index(wred,w[i]);
		list := toms[position];
		Append(~list, "Tom(" cat IntegerToString(i) cat ")");
		toms[position] := list;
	end for;
	// Detect equivalent Jerry formats
	// Record inequivalent first cases we've found so far as
	//	[[w-entry1,w-entry2], ... ]
	so_far := [ ];
	jerries := [ Parent(["Jerry(3,5)","Jerry(4,5)"]) | ];
	all_jerries := [ [i,j] : j in [i+1..5], i in [1..4] ];
	jerries_in_w := [ [w[I[1]],w[I[2]]] : I in all_jerries ];
	for i in [1..#all_jerries] do
		this_type := all_jerries[i];
		this_type_str := "Jerry(" cat IntegerToString(this_type[1])
			    cat "," cat IntegerToString(this_type[2]) cat ")";
		this_one := jerries_in_w[i];
		seen := Index(so_far,this_one);
		if seen ne 0 then	// i.e. seen an equivalent one already
			list := jerries[seen];	// this must exist
			Append(~list,this_type_str);
			jerries[seen] := list;
		else
			Append(~jerries,[this_type_str]);
			Append(~so_far,this_one);
		end if;
	end for;
	return toms cat jerries;
end function;

// return entries [i,j], i<j, in M that equal 0.
function fishy_zero(M)
	return [Parent([1,2])| [i,j] :j in [i+1..5],i in [1..4] | M[i,j] eq 0];
end function;

